@@ -22,15 +22,55 @@ import re
 from . import ast
 from . import message
 from .annotationparser import (TAG_DEPRECATED, TAG_SINCE, TAG_STABILITY, TAG_RETURNS)
-from .annotationparser import (ANN_ALLOW_NONE, ANN_ARRAY, ANN_ATTRIBUTES, ANN_CLOSURE,
-                               ANN_CONSTRUCTOR, ANN_DESTROY, ANN_ELEMENT_TYPE, ANN_FOREIGN,
-                               ANN_GET_VALUE_FUNC, ANN_IN, ANN_INOUT, ANN_METHOD, ANN_OUT,
-                               ANN_REF_FUNC, ANN_RENAME_TO, ANN_SCOPE, ANN_SET_VALUE_FUNC,
-                               ANN_SKIP, ANN_TRANSFER, ANN_TYPE, ANN_UNREF_FUNC, ANN_VALUE,
-                               ANN_VFUNC, ANN_NULLABLE, ANN_OPTIONAL, ANN_NOT)
-from .annotationparser import (OPT_ARRAY_FIXED_SIZE, OPT_ARRAY_LENGTH, OPT_ARRAY_ZERO_TERMINATED,
-                               OPT_OUT_CALLEE_ALLOCATES, OPT_OUT_CALLER_ALLOCATES,
-                               OPT_TRANSFER_CONTAINER, OPT_TRANSFER_FLOATING, OPT_TRANSFER_NONE)
+from .annotationparser import (
+    ANN_ALLOW_NONE,
+    ANN_ARRAY,
+    ANN_ATTRIBUTES,
+    ANN_CLOSURE,
+    ANN_CONSTRUCTOR,
+    ANN_COPY_FUNC,
+    ANN_DEFAULT_VALUE,
+    ANN_DESTROY,
+    ANN_ELEMENT_TYPE,
+    ANN_EMITTER,
+    ANN_FOREIGN,
+    ANN_FREE_FUNC,
+    ANN_GET_PROPERTY,
+    ANN_GET_VALUE_FUNC,
+    ANN_GETTER,
+    ANN_IN,
+    ANN_INOUT,
+    ANN_METHOD,
+    ANN_OUT,
+    ANN_REF_FUNC,
+    ANN_RENAME_TO,
+    ANN_SCOPE,
+    ANN_SET_PROPERTY,
+    ANN_SET_VALUE_FUNC,
+    ANN_SETTER,
+    ANN_SKIP,
+    ANN_TRANSFER,
+    ANN_TYPE,
+    ANN_UNREF_FUNC,
+    ANN_VALUE,
+    ANN_VFUNC,
+    ANN_NULLABLE,
+    ANN_OPTIONAL,
+    ANN_NOT,
+    ANN_FINISH_FUNC,
+    ANN_SYNC_FUNC,
+    ANN_ASYNC_FUNC,
+)
+from .annotationparser import (
+    OPT_ARRAY_FIXED_SIZE,
+    OPT_ARRAY_LENGTH,
+    OPT_ARRAY_ZERO_TERMINATED,
+    OPT_OUT_CALLEE_ALLOCATES,
+    OPT_OUT_CALLER_ALLOCATES,
+    OPT_TRANSFER_CONTAINER,
+    OPT_TRANSFER_FLOATING,
+    OPT_TRANSFER_NONE,
+)
 
 from .utils import to_underscores_noprefix
 
@@ -91,6 +131,7 @@ class MainTransformer(object):
                 self._pair_function(node)
             if isinstance(node, (ast.Class, ast.Interface)):
                 self._pair_class_virtuals(node)
+                self._pair_property_accessors(node)
 
         # Some annotations need to be post function pairing
         self._namespace.walk(self._pass_read_annotations2)
@@ -274,6 +315,13 @@ class MainTransformer(object):
                 node.get_value_func = annotation[0] if annotation else None
         if isinstance(node, ast.Constant):
             self._apply_annotations_constant(node)
+        if isinstance(node, (ast.Record, ast.Union)):
+            block = self._get_block(node)
+            if block:
+                annotation = block.annotations.get(ANN_COPY_FUNC)
+                node.copy_func = annotation[0] if annotation else None
+                annotation = block.annotations.get(ANN_FREE_FUNC)
+                node.free_func = annotation[0] if annotation else None
         return True
 
     def _adjust_container_type(self, parent, node, annotations):
@@ -595,9 +643,10 @@ class MainTransformer(object):
             transfer = OPT_TRANSFER_NONE
 
             if (not isinstance(target, (ast.Class, ast.Interface))
-                    and node_type.target_giname != 'GLib.Variant'):
+                    and node_type.target_giname != 'GLib.Variant'
+                    and node_type.target_giname != 'GObject.Closure'):
                 message.warn('invalid "transfer" annotation for {0}: '
-                             'only valid for object and GVariant types'.format(target),
+                             'only valid for object, GVariant and GClosure types'.format(target),
                              annotations.position)
                 return
 
@@ -773,6 +822,14 @@ class MainTransformer(object):
         if ANN_METHOD in block.annotations:
             node.is_method = True
 
+        set_property = block.annotations.get(ANN_SET_PROPERTY)
+        if set_property is not None and isinstance(node, ast.Function):
+            node.set_property = set_property[0]
+
+        get_property = block.annotations.get(ANN_GET_PROPERTY)
+        if get_property is not None and isinstance(node, ast.Function):
+            node.get_property = get_property[0]
+
     def _apply_annotations_alias(self, node, chain):
         block = self._get_block(node)
         self._apply_annotations_annotated(node, block)
@@ -868,6 +925,18 @@ class MainTransformer(object):
                 param.position)
 
     def _apply_annotations_callable(self, node, chain, block):
+        if block is not None:
+            finish_func = block.annotations.get(ANN_FINISH_FUNC)
+            if finish_func is not None and isinstance(node, ast.Callable):
+                node.finish_func = finish_func[0]
+
+            sync_func = block.annotations.get(ANN_SYNC_FUNC)
+            if sync_func is not None and isinstance(node, ast.Callable):
+                node.sync_func = sync_func[0]
+
+            async_func = block.annotations.get(ANN_ASYNC_FUNC)
+            if async_func is not None and isinstance(node, ast.Callable):
+                node.async_func = async_func[0]
         self._apply_annotations_annotated(node, block)
         self._apply_annotations_params(node, node.parameters, block)
         self._apply_annotations_return(node, node.retval, block)
@@ -914,6 +983,15 @@ class MainTransformer(object):
         type_annotation = block.annotations.get(ANN_TYPE)
         if type_annotation:
             prop.type = self._resolve_toplevel(type_annotation[0], prop.type, prop, parent)
+        setter = block.annotations.get(ANN_SETTER)
+        if setter:
+            prop.setter = setter[0]
+        getter = block.annotations.get(ANN_GETTER)
+        if getter:
+            prop.getter = getter[0]
+        default_value = block.annotations.get(ANN_DEFAULT_VALUE)
+        if default_value:
+            prop.default_value = default_value[0]
 
     def _apply_annotations_signal(self, parent, signal):
         names = []
@@ -922,6 +1000,11 @@ class MainTransformer(object):
 
         if block:
             self._apply_annotations_annotated(signal, block)
+
+            emitter = block.annotations.get(ANN_EMITTER)
+            if emitter:
+                signal.emitter = emitter[0]
+
             # We're only attempting to name the signal parameters if
             # the number of parameters (@foo) is the same or greater
             # than the number of signal parameters
@@ -1288,10 +1371,21 @@ method or constructor of some type."""
         return origin_node
 
     def _get_constructor_name(self, func, subsymbol):
-        name = None
+        prefix_matches = False
+        uscored_prefix = None
+        target = self._transformer.lookup_typenode(func.retval.type)
+        if hasattr(target, 'c_symbol_prefix') and target.c_symbol_prefix is not None:
+            prefix_matches = subsymbol.startswith(target.c_symbol_prefix)
+            if prefix_matches:
+                uscored_prefix = target.c_symbol_prefix
+        if not prefix_matches:
+            uscored_prefix = self._uscored_identifier_for_type(func.retval.type)
         split = self._split_uscored_by_type(subsymbol)
         if split is None:
             if func.is_constructor:
+                if uscored_prefix in func.symbol:
+                    subsym_idx = func.symbol.find(subsymbol)
+                    func.name = func.symbol[(subsym_idx + len(uscored_prefix) + 1):]
                 name = func.name
         else:
             _, name = split
@@ -1420,7 +1514,12 @@ method or constructor of some type."""
             vfunc.inherit_file_positions(callback)
 
             prefix = self._get_annotation_name(class_struct)
+            # Prefer full docblocks, but fall back to the field description
+            # if there isn't a full one, to avoid an undocumented symbol
             block = self._blocks.get('%s::%s' % (prefix, vfunc.name))
+            if block is None:
+                vfunc.doc = field.doc
+                vfunc.doc_position = field.doc_position
             self._apply_annotations_callable(vfunc, [node], block)
             node.virtual_methods.append(vfunc)
 
@@ -1447,18 +1546,215 @@ method or constructor of some type."""
                 self._apply_annotations_callable(vfunc, [], block)
                 break
 
+    def _pair_property_accessors(self, node):
+        """Look for accessor methods for class properties"""
+        for prop in node.properties:
+            if not prop.introspectable:
+                continue
+            if prop.setter is None:
+                normalized_name = prop.name.replace('-', '_')
+                if prop.writable and not prop.construct_only:
+                    setter = 'set_' + normalized_name
+                else:
+                    setter = None
+            else:
+                setter = prop.setter
+            if prop.getter is None:
+                if prop.readable:
+                    getter = ['get_' + normalized_name]
+                    # Heuristic: read-only properties can have getters that are
+                    # just the property name, like: gtk_widget_has_focus()
+                    if not prop.writable and prop.type.is_equiv(ast.TYPE_BOOLEAN):
+                        getter.append(normalized_name)
+                else:
+                    getter = []
+            else:
+                getter = [prop.getter]
+            for method in node.methods:
+                if not method.introspectable:
+                    continue
+                if setter is not None and method.name == setter:
+                    if method.set_property is None:
+                        method.set_property = prop.name
+                    elif method.set_property != prop.name:
+                        message.warn_node(method,
+                                          "Setter method '%s' for property '%s' has a "
+                                          "mismatched '(set-property %s)' annotation" %
+                                          (method.symbol, prop.name, method.set_property))
+                        method.set_property = prop.name
+                    prop.setter = method.name
+                    continue
+                if getter is not [] and method.name in getter:
+                    if method.get_property is None:
+                        method.get_property = prop.name
+                    elif method.get_property != prop.name:
+                        message.warn_node(method,
+                                          "Getter method '%s' for property '%s' has a "
+                                          "mismatched '(get-property %s)' annotation" %
+                                          (method.symbol, prop.name, method.get_property))
+                        method.get_property = prop.name
+                    prop.getter = method.name
+                    continue
+
     def _pass3(self, node, chain):
         """Pass 3 is after we've loaded GType data and performed type
         closure."""
         if isinstance(node, ast.Callable):
             self._pass3_callable_callbacks(node)
             self._pass3_callable_throws(node)
+            self._pass3_callable_async_finish(node)
+            self._pass3_callable_async_sync(node)
+        elif isinstance(node, (ast.Class, ast.Interface)):
+            self._pass3_class_async_finish(node)
+            self._pass3_class_async_sync(node)
         return True
 
+    def _pass3_class_async_finish(self, node):
+        self._match_class_async_methods(node.methods)
+        self._match_class_async_methods(node.virtual_methods)
+        self._match_class_async_methods(node.static_methods + node.constructors)
+
+    def _pass3_class_async_sync(self, node):
+        self._match_class_sync_methods(node.methods)
+        self._match_class_sync_methods(node.virtual_methods)
+        self._match_class_sync_methods(node.static_methods)
+
+    def _pass3_callable_async_finish(self, node):
+        if node.finish_func is not None:
+            return
+        for param in node.parameters:
+            if param.type.ctype is None or param.type.ctype not in ('GAsyncReadyCallback'):
+                continue
+            func_name = node.name + '_finish'
+            if (node.name.endswith('_async')):
+                func_name = node.name[:-6] + '_finish'
+            if node.parent is None:
+                continue
+            if ('GAsyncResult') in node.parent.ctypes:
+                if func_name in node.parent:
+                    node.finish_func = func_name
+                    break
+                else:
+                    message.warn_node(node,
+                    "Couldn't find '%s' for the corresponding async function: '%s'"
+                    % (func_name, node.name))
+
+    def _pass3_callable_async_sync(self, node):
+        if node.sync_func is not None or node.finish_func is None:
+            return
+        if (node.name.endswith('_async')):
+            sync_name = node.name[:-6]
+        else:
+            sync_name = node.name + '_sync'
+        if node.parent is None:
+            return
+        finish_func = node.parent.get(node.finish_func)
+        if finish_func is None:
+            return
+        candidate_method = node.parent.get(sync_name)
+        if candidate_method is None:
+            return
+        if finish_func.retval.type.ctype != candidate_method.retval.type.ctype:
+            return
+        param_matched = True
+        for param in candidate_method.parameters:
+            if param.direction == ast.PARAM_DIRECTION_IN:
+                for candidate_param in node.parameters:
+                    if (candidate_param.direction == ast.PARAM_DIRECTION_IN and
+                            candidate_param.name == param.name):
+                        break
+                else:
+                    param_matched = False
+                    break
+            elif param.direction == ast.PARAM_DIRECTION_OUT:
+                for candidate_param in finish_func.parameters:
+                    if (candidate_param.direction == ast.PARAM_DIRECTION_OUT and
+                            candidate_param.name == param.name):
+                        break
+                else:
+                    param_matched = False
+                    break
+        if param_matched:
+            node.sync_func = candidate_method.name
+            candidate_method.async_func = node.name
+            return
+
+    def _match_class_async_methods(self, methods):
+        for method in methods:
+            if method.finish_func is not None:
+                continue
+            func_name = method.name + '_finish'
+            if method.name.endswith('_async'):
+                func_name = method.name[:-6] + '_finish'
+            found_async_callback_param = False
+            for params in method.parameters:
+                if (params.type.ctype == 'GAsyncReadyCallback'):
+                    found_async_callback_param = True
+
+            if not found_async_callback_param:
+                continue
+
+            found_finish_method = False
+            for candidate_method in methods:
+                if found_finish_method:
+                    break
+                if candidate_method.name != func_name:
+                    continue
+                for candidate_param in candidate_method.parameters:
+                    if candidate_param.type.ctype == 'GAsyncResult*':
+                        method.finish_func = candidate_method.name
+                        found_finish_method = True
+                        break
+            if not found_finish_method:
+                message.warn_node(method,
+                "Couldn't find '%s' for the corresponding async function: '%s'"
+                % (func_name, method.name))
+
+    def _match_class_sync_methods(self, methods):
+        for method in methods:
+            if method.sync_func is not None or method.finish_func is None:
+                continue
+            if method.name.endswith('_async'):
+                sync_name = method.name[:-6]
+            else:
+                sync_name = method.name + '_sync'
+            param_matched = True
+            for finish_candidate_method in methods:
+                type_name = finish_candidate_method.retval.type.ctype
+                if finish_candidate_method.name == method.finish_func:
+                    type_name = finish_candidate_method.retval.type.ctype
+                    break
+            for candidate_method in methods:
+                if candidate_method.name != sync_name:
+                    continue
+                if type_name != candidate_method.retval.type.ctype:
+                    param_matched = False
+                    break
+                for param in candidate_method.parameters:
+                    if param.direction == ast.PARAM_DIRECTION_IN:
+                        for candidate_param in method.parameters:
+                            if (candidate_param.direction == ast.PARAM_DIRECTION_IN and
+                                    candidate_param.name == param.name):
+                                break
+                        else:
+                            param_matched = False
+                            break
+                    elif param.direction == ast.PARAM_DIRECTION_OUT:
+                        for candidate_param in finish_candidate_method.parameters:
+                            if (candidate_param.direction == ast.PARAM_DIRECTION_OUT and
+                                    candidate_param.name == param.name):
+                                break
+                        else:
+                            param_matched = False
+                            break
+                if param_matched:
+                    method.sync_func = candidate_method.name
+                    candidate_method.async_func = method.name
+                    break
+
     def _pass3_callable_callbacks(self, node):
         """Check to see if we have anything that looks like a
         callback+user_data+GDestroyNotify set."""
-
         params = node.parameters
 
         # First, do defaults for well-known callback types
